#include "GLee.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <iostream>
#include <fstream>
#include "globals.h"

using namespace std;

//window resize callback
void resizeWindow(int w, int h)
{
	// Set the OpenGL viewport to fill the entire window
	glViewport(0, 0, (GLsizei)w, (GLsizei)h);
	
	s_width = w;
	s_height = h;
	camera->glInit(s_width,s_height);
}

//keyboard event callback
void keyboardEvent(unsigned char key, int x, int y)
{
	switch (key) {
	case 'q':  case 'Q':
		exit(0);
		break;
		
	case 'a': case 'A':
		ssao_toggle = !ssao_toggle;
		glutPostRedisplay();
		break;
		
	case 'b': case 'B':
		blur_toggle = !blur_toggle;
		glutPostRedisplay();
		break;
		
	case 'o': case 'O':
		outline_toggle = !outline_toggle;
		glutPostRedisplay();
		break;
		
	case 'c': case 'C':
		color_toggle = !color_toggle;
		glutPostRedisplay();
		break;
		
	case 't': case'T':
		toon_toggle = !toon_toggle;
		glutPostRedisplay();
		break;
		
	case 'l': case'L':
		lighting_toggle = !lighting_toggle;
		glutPostRedisplay();
		break;
		
	default:
	cerr << "UNKNOWN KEYBOARD INPUT  '"<< key <<"'" << endl;
	}
}

//mouse event callback
void mouseEvent(int button, int state, int x, int y)
{
	// Save the current state of the mouse.  This will be
	// used by the 'motion' function
	mouseButton = button;
	mouseX = x;
	mouseY = y;
	controlPressed = glutGetModifiers() & GLUT_ACTIVE_CTRL;
}


//mouse motion callback
void mouseMotion(int x, int y)
{
	// Left button = rotation
	// (rotate camera around the up and horizontal vectors)
	if (mouseButton == GLUT_LEFT_BUTTON) {
		camera->rotateCamera(0.005*(mouseX-x), 0.005*(mouseY-y));
		mouseX = x;
		mouseY = y;
	}
	// Middle button = translation
	// (move camera perpendicular to the direction vector)
	else if (mouseButton == GLUT_MIDDLE_BUTTON) {
		camera->truckCamera((mouseX-x)*0.5, (y-mouseY)*0.5);
		mouseX = x;
		mouseY = y;
	}
	// Right button = dolly or zoom
	// (move camera along the direction vector)
	else if (mouseButton == GLUT_RIGHT_BUTTON) {
		if (controlPressed) {
			camera->zoomCamera(mouseY-y);
		} else {
			camera->dollyCamera(mouseY-y);
		}
		mouseX = x;
		mouseY = y;
	}

	// Redraw the scene with the new camera parameters
	glutPostRedisplay();
}


//load and compile a shader
GLuint loadShader(GLenum s_type, string filename)
{
	ifstream file(filename.c_str());
	if(!file)
	{
		cerr << "ERROR: Shader not found: " << filename << endl;
		exit(1);
	}
	
	//read shader into a string
	string shadertext;
	string tmp;
	while(getline(file,tmp))
		shadertext += tmp + '\n';
	
	const GLchar *sourcePtr[1];
	sourcePtr[0] = shadertext.c_str();

	//create shader handle
	GLuint handle = glCreateShader(s_type);

	//load shader source into GL
	glShaderSource(handle, 1, sourcePtr, NULL);
	
	//compile shader
	glCompileShader(handle);
	
	GLint success = 0;
	glGetShaderiv(handle, GL_COMPILE_STATUS, &success);
	if (!success)
    {
        GLchar infoLog[1024];
        glGetShaderInfoLog(handle, 1024, NULL, infoLog);
        cerr << "Error in fragment shader compilation!" << endl;
        cerr << "Info log: " << infoLog << endl;
        exit(0);
    }
    
	return handle;
}


//link two shaders into a program
GLuint linkShader(GLuint shader1, GLuint shader2)
{
	GLuint handle = glCreateProgram();
	glAttachShader(handle, shader1);
	glAttachShader(handle, shader2);
	glLinkProgram(handle);
	
	GLint success = 0;
	glGetProgramiv(handle, GL_LINK_STATUS, &success);
    if (!success)
    {
        GLchar infoLog[1024];
        glGetProgramInfoLog(handle, 1024, NULL, infoLog);
        cerr << "Error in program linkage!" << endl;
        cerr << "Info log: " << infoLog << endl;
        exit(0);
    }
    
    success = 0;
    glValidateProgram(handle);
    glGetProgramiv(handle, GL_VALIDATE_STATUS, &success);
    if (!success)
    {
    	GLchar infoLog[1024];
        glGetProgramInfoLog(handle, 1024, NULL, infoLog);
        cerr << "Error in program validation!" << endl;
        cerr << "Info log: " << infoLog << endl;
        exit(0);
    }
	
	return handle;
}
