#include "GLee.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <iostream>
#include <fstream>
#include <cmath>

#include "globals.h"
#include "argparser.h"
#include "camera.h"
#include "mesh.h"
#include "funcs.h"

using namespace std;

//textures for rendering to
GLuint fbo;
GLuint normal_tex;
GLuint depth_tex;

GLuint ssao_fbo;
GLuint ssao_tex;
GLuint blur_fbo;
GLuint blur_tex;

//random normal texture
GLuint rand_norm;

//uniforms for shaders
GLint uniformLoc;
GLfloat ssao_samples[24];
GLfloat lightPosition[3];

//shaders
GLuint n_shader, n_shader_v, n_shader_f;
GLuint ssao_v, ssao_f, ssao;
GLuint toon_v, toon_f, toon;
GLuint blur_v, blur_f, blur;
GLuint specular_v, specular_f, specular;

//glut display callback
//this is where rendering takes place
void display()
{
	//
	//1st pass:
	//  render normal/depth texture
	//
	glDisable(GL_LIGHTING);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, s_width, s_height);

	camera->glInit(s_width,s_height);
	glClearColor(1.0,1.0,1.0,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	camera->glPlaceCamera();
	
	glUseProgram(n_shader);
	scene_mesh->render();

	glPopAttrib();
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);


	//
	//2nd pass:
	//  render ssao to texture
	//
	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, ssao_fbo);
	
	glClearColor(1.0,1.0,1.0,0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	if(ssao_toggle){
	
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, depth_tex);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, rand_norm);
		
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	
		glUseProgram(ssao);
		uniformLoc = glGetUniformLocation(ssao, "s_depth");
		if (uniformLoc != -1)
		{
			glUniform1i(uniformLoc, 0);
		}
	
		uniformLoc = glGetUniformLocation(ssao, "s_norm");
		if (uniformLoc != -1)
		{
			glUniform1i(uniformLoc, 1);
		}
		
		uniformLoc = glGetUniformLocation(ssao, "sample_points");
		if (uniformLoc != -1)
		{
			glUniform3fv(uniformLoc, 8, ssao_samples);
		}
		
		//render texture to fullscreen quad
		glBegin(GL_QUADS);
			glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 0.0f, 0.0f);
			glVertex3f(0.0f, 0.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 8.0f, 0.0f);
			glVertex3f(1.0f, 0.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 1.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 8.0f, 8.0f);
			glVertex3f(1.0f, 1.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 1.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 0.0f, 8.0f);
			glVertex3f(0.0f, 1.0f, 0.0f);
		glEnd();
	}
	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	

	
	//
	//3rd pass: blur!
	//
	if(blur_toggle)
	{
		for(int bb=0; bb < 5; bb++)
		{
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, blur_fbo);
			
			glClear(GL_COLOR_BUFFER_BIT);
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrtho(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			
			glActiveTexture(GL_TEXTURE0);
			glBindTexture(GL_TEXTURE_2D, ssao_tex);
			
			glUseProgram(blur);
			
			uniformLoc = glGetUniformLocation(blur, "sampler0");
			if (uniformLoc != -1)
			{
				glUniform1i(uniformLoc, 0);
			}
			
			glColor3f(1.0,1.0,1.0);
			glBegin(GL_QUADS);
				glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 0.0f);
				glVertex3f(0.0f, 0.0f, 0.0f);
				glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 0.0f);
				glVertex3f(1.0f, 0.0f, 0.0f);
				glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 1.0f);
				glVertex3f(1.0f, 1.0f, 0.0f);
				glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 1.0f);
				glVertex3f(0.0f, 1.0f, 0.0f);
			glEnd();
			
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
			
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, ssao_fbo);
			
			glClear(GL_COLOR_BUFFER_BIT);
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrtho(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			
			glActiveTexture(GL_TEXTURE0);
			glBindTexture(GL_TEXTURE_2D, blur_tex);
			
			glUseProgram(blur);
			
			uniformLoc = glGetUniformLocation(blur, "sampler0");
			if (uniformLoc != -1)
			{
				glUniform1i(uniformLoc, 0);
			}
			
			glColor3f(1.0,1.0,1.0);
			glBegin(GL_QUADS);
				glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 0.0f);
				glVertex3f(0.0f, 0.0f, 0.0f);
				glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 0.0f);
				glVertex3f(1.0f, 0.0f, 0.0f);
				glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 1.0f);
				glVertex3f(1.0f, 1.0f, 0.0f);
				glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 1.0f);
				glVertex3f(0.0f, 1.0f, 0.0f);
			glEnd();
			
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		}
	}
	//
	//4th Pass: Actual Shading
	//
	
	// Clear the display buffer, set it to the background color
	if (color_toggle)
	{
		glClearColor(0.2, 0.6, 0.9, 0);
	}
	else
	{
		glClearColor(1.0, 1.0, 1.0, 0);
	}
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//camera
	camera->glInit(s_width,s_height);
	glEnable(GL_DEPTH_TEST);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	camera->glPlaceCamera();

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, ssao_tex);

	//render the scene with the specular shader
	glUseProgram(specular);
	
	uniformLoc = glGetUniformLocation(specular, "sampler0");
	if (uniformLoc != -1)
	{
		glUniform1i(uniformLoc, 0);
	}
	
	uniformLoc = glGetUniformLocation(specular, "lightPos");
	if (uniformLoc != -1)
	{
		glUniform3fv(uniformLoc, 1, lightPosition);
	}
	
	uniformLoc = glGetUniformLocation(specular, "color_toggle");
	if (uniformLoc != -1)
	{
		glUniform1i(uniformLoc, color_toggle);
	}
	
	uniformLoc = glGetUniformLocation(specular, "toon_toggle");
	if (uniformLoc != -1)
	{
		glUniform1i(uniformLoc, toon_toggle);
	}
	
	uniformLoc = glGetUniformLocation(specular, "lighting_toggle");
	if (uniformLoc != -1)
	{
		glUniform1i(uniformLoc, lighting_toggle);
	}

	scene_mesh->render();
	

	
	//
	//5th Pass: Outlines
	//
	if(outline_toggle)
	{
		glEnable(GL_BLEND);
		glBlendFunc(GL_DST_COLOR, GL_ZERO);
		
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, depth_tex);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, normal_tex);
		
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
	
		glUseProgram(toon);
		uniformLoc = glGetUniformLocation(toon, "s_depth");
		if (uniformLoc != -1)
		{
			glUniform1i(uniformLoc, 0);
		}
	
		uniformLoc = glGetUniformLocation(toon, "s_norm");
		if (uniformLoc != -1)
		{
			glUniform1i(uniformLoc, 1);
		}
		
		//render texture to fullscreen quad
		glBegin(GL_QUADS);
			glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 0.0f, 0.0f);
			glVertex3f(0.0f, 0.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 8.0f, 0.0f);
			glVertex3f(1.0f, 0.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 1.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 8.0f, 8.0f);
			glVertex3f(1.0f, 1.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 1.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 0.0f, 8.0f);
			glVertex3f(0.0f, 1.0f, 0.0f);
		glEnd();
		
		glDisable(GL_BLEND);
	}
	//swap buffers. yay!
	glutSwapBuffers();
}



//intialize GL
void init()
{
	//camera
	Vec3f point_of_interest = scene_mesh->getBoundingBox()->getCenter();
	float max_dim = scene_mesh->getBoundingBox()->maxDim();
	Vec3f camera_position = point_of_interest + Vec3f(0,0,4*max_dim);
	Vec3f up = Vec3f(0,1,0);
	camera = new PerspectiveCamera(camera_position, point_of_interest, up, 20 * M_PI/180.0);
	camera->glInit(s_width,s_height);
	
	//lighting, also yay!
	glEnable(GL_LIGHTING);
	glShadeModel(GL_SMOOTH);

	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGBA);
	glEnable(GL_DEPTH_TEST);
	glutInitWindowSize(s_width, s_height);
	glutInitWindowPosition(100,100);
	glutCreateWindow("NPR SSAO");
	
	glEnable(GL_NORMALIZE);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_2D);
	
	//Initialize Lighting
	glEnable(GL_COLOR_MATERIAL);
	GLfloat lightPos[] = {2.0f, 2.0f, 2.0f, 1.0f};
	GLfloat ambientLight[] = {0.3f, 0.3f, 0.3f, 1.0f};
	GLfloat diffuseLight[] = {0.7f, 0.7f, 0.7f, 1.0f};
	GLfloat specularLight[] = {1.0f,1.0f,1.0f,1.0f};
	
	glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
	glLightfv(GL_LIGHT0,GL_AMBIENT,ambientLight);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuseLight);
	glLightfv(GL_LIGHT0,GL_SPECULAR,specularLight);
	glEnable(GL_LIGHT0);
	
	lightPosition[0] = 5.0;
	lightPosition[1] = 5.0;
	lightPosition[2] = 0.0;
	
	//glut callbacks
	glutMouseFunc(mouseEvent);
	glutMotionFunc(mouseMotion);
	glutReshapeFunc(resizeWindow);
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboardEvent);
	
	
	//check for shader support
	if (!GLEE_VERSION_2_0 && (!GLEE_ARB_fragment_shader || 
                              !GLEE_ARB_shader_objects || 
                              !GLEE_ARB_shading_language_100))
    {
        cerr << "ERROR: GLSL extensions not available!" << endl;
        exit(1);
    }
    
    //make textures
    glGenFramebuffersEXT(1, &fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);

	glGenTextures(1, &normal_tex);
	glBindTexture(GL_TEXTURE_2D, normal_tex);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, s_width, s_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    
    glGenTextures(1, &depth_tex);
	glBindTexture(GL_TEXTURE_2D, depth_tex);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT16_ARB, s_width, s_height, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, NULL);  
    
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, normal_tex, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depth_tex, 0);
	
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if(status != GL_FRAMEBUFFER_COMPLETE_EXT)
	{
		cerr << "ERROR: Framebuffer creation failed." << endl;
		exit(1);
	}
	
	//make ssao fbo and textures
	glGenFramebuffersEXT(1, &ssao_fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, ssao_fbo);
	
	glGenTextures(1, &ssao_tex);
	glBindTexture(GL_TEXTURE_2D, ssao_tex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, s_width, s_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, ssao_tex, 0);
	
	status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if(status != GL_FRAMEBUFFER_COMPLETE_EXT)
	{
		cerr << "ERROR: Framebuffer creation failed." << status << endl;
		exit(1);
	}
	
	//make blur fbo and textures
	glGenFramebuffersEXT(1, &blur_fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, blur_fbo);
	
	glGenTextures(1, &blur_tex);
	glBindTexture(GL_TEXTURE_2D, blur_tex);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, s_width, s_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    	
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, blur_tex, 0);
	
	status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if(status != GL_FRAMEBUFFER_COMPLETE_EXT)
	{
		cerr << "ERROR: Framebuffer creation failed: " << endl;
		exit(1);
	}
	
	
	//ssao samples - uniform sampling
	ssao_samples[0] = 1.0; ssao_samples[1] = 1.0; ssao_samples[2] = 1.0;
	ssao_samples[3] = -1.0; ssao_samples[4] = 1.0; ssao_samples[5] = 1.0;
	ssao_samples[6] = -1.0; ssao_samples[7] = -1.0; ssao_samples[8] = 1.0;
	ssao_samples[9] = -1.0; ssao_samples[10] = 1.0; ssao_samples[11] = -1.0;
	ssao_samples[12] = -1.0; ssao_samples[13] = -1.0; ssao_samples[14] = -1.0;
	ssao_samples[15] = 1.0; ssao_samples[16] = -1.0; ssao_samples[17] = -1.0;
	ssao_samples[18] = 1.0; ssao_samples[19] = 1.0; ssao_samples[20] = -1.0;
	ssao_samples[21] = 1.0; ssao_samples[22] = -1.0; ssao_samples[23] = 1.0;
	
	//normalize the samples
	for(int i=0; i < 24; i=i+4)
	{
		float len = pow(ssao_samples[i],2) + pow(ssao_samples[i+1],2) + pow(ssao_samples[i+2],2);
		len = sqrt(len); 
		ssao_samples[i] /= len;
		ssao_samples[i+1] /= len;
		ssao_samples[i+2] /= len;
	}
	
	//create the random normal texture
	srand(time(NULL));
	GLbyte *rn_data = new GLbyte[16384];
	for(int i = 0; i < 16384; i = i + 4)
	{
		float r = (float)rand() / (float)RAND_MAX;
		float g = (float)rand() / (float)RAND_MAX;
		float b = (float)rand() / (float)RAND_MAX;
		float a = (float)rand() / (float)RAND_MAX;
		float len = sqrt(r*r + g*g + b*b + a*a);
		r /= len; g /= len; b /= len; a /= len;
		rn_data[i] = 255 * r;
		rn_data[i+1] = 255 * g;
		rn_data[i+2] = 255 * b;
		rn_data[i+3] = 255 * a;
	}
	
	glGenTextures(1, &rand_norm);
	glBindTexture(GL_TEXTURE_2D, rand_norm);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, 64, 64, 0, GL_RGBA, GL_BYTE, rn_data);
    delete [] rn_data;  
    
}


int main(int argc, char* argv[])
{
	ssao_toggle = true;
	blur_toggle = true;
	outline_toggle = true;
	toon_toggle = true;
	color_toggle = true;
	lighting_toggle = true;
	
	//intialize GLUT
	glutInit(&argc,argv);
	
	//get arguments and load mesh
	ArgParser args(argc, argv);
	s_width = 512; s_height = 512;
	
	scene_mesh = new Mesh(args.input_file);
	
	//intialize OpenGL/GLUT and start rendering
	init();
	
	//load shaders	
	n_shader_v = loadShader(GL_VERTEX_SHADER, "shaders/normals.vs");
	n_shader_f = loadShader(GL_FRAGMENT_SHADER, "shaders/normals.fs");
	n_shader = linkShader(n_shader_v,n_shader_f);
	
	ssao_v = loadShader(GL_VERTEX_SHADER, "shaders/ssao.vs");
	ssao_f = loadShader(GL_FRAGMENT_SHADER, "shaders/ssao.fs");
	ssao = linkShader(ssao_v,ssao_f); 
	
	toon_v = loadShader(GL_VERTEX_SHADER, "shaders/toon_outlining.vs");
	toon_f = loadShader(GL_FRAGMENT_SHADER, "shaders/toon_outlining.fs");
	toon = linkShader(toon_v, toon_f); 
	
	blur_v = loadShader(GL_VERTEX_SHADER, "shaders/blur.vs");
	blur_f = loadShader(GL_FRAGMENT_SHADER, "shaders/blur.fs");
	blur = linkShader(blur_v,blur_f); 
	
	specular_v = loadShader(GL_VERTEX_SHADER, "shaders/specular.vs");
	specular_f = loadShader(GL_FRAGMENT_SHADER, "shaders/specular.fs");
	specular = linkShader(specular_v,specular_f); 
	
	glutMainLoop();
	
	//deallocate shaders
	glDeleteProgram(n_shader);
	glDeleteShader(n_shader_v);
	glDeleteShader(n_shader_f);
	
	glDeleteProgram(ssao);
	glDeleteShader(ssao_f);
	glDeleteShader(ssao_v);
	
	glDeleteProgram(toon);
	glDeleteShader(toon_f);
	glDeleteShader(toon_v);
	
	glDeleteProgram(blur);
	glDeleteShader(blur_f);
	glDeleteShader(blur_v);
	
	//cleanup everything else
	glDeleteFramebuffersEXT(1, &fbo);
	glDeleteFramebuffersEXT(1, &ssao_fbo);
	glDeleteFramebuffersEXT(1, &blur_fbo);
	
	glDeleteTextures(1, &normal_tex);
	glDeleteTextures(1, &depth_tex);
	glDeleteTextures(1, &ssao_tex);
	glDeleteTextures(1, &blur_tex);
	glDeleteTextures(1, &rand_norm);
		
	return 0;
}
