#ifndef _MATERIAL_H_
#define _MATERIAL_H_

#include <GL/gl.h>
#include <GL/glu.h>
#include <string>
#include <iostream>
#include "vectors.h"

using namespace std;
class Material
{
public:
	Material(): name("null"), ns(96.0), ambient(Vec3f(0.0,0.0,0.0)), diffuse(Vec3f(0.8,0.8,0.8)), specular(Vec3f(1.0,1.0,1.0)){};
	Material(string n, float Ns, Vec3f Ka, Vec3f Kd, Vec3f Ks): name(n), ns(Ns), ambient(Ka), diffuse(Kd), specular(Ks){};

	void setName(string n){ name = n; }
	void setExponent(float se){ ns = se; }
	void setAmbient(Vec3f a){ ambient = a; }
	void setDiffuse(Vec3f d){ diffuse = d; }
	void setSpecular(Vec3f s){ specular = s; }
	
	//outputs OpenGL commands for changing the material
	void gl()
	{
		//diffuse
		glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);
		glColor3f(diffuse.x(),diffuse.y(),diffuse.z());
		
		//specular
		GLfloat spec[] = {specular.x(),specular.y(),specular.z(),1.0f};
		glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
		
		//specular exponent
		glMaterialf(GL_FRONT, GL_SHININESS, ns);
	}
	
private:
	std::string name;
	float ns; //specular exponent
	Vec3f ambient;
	Vec3f diffuse;
	Vec3f specular;
};

#endif
