#ifndef _MESH_H_
#define _MESH_H_

#include "boundingbox.h"
#include "vectors.h"
#include "material.h"
#include <vector>
#include <map>
#include <pcre++.h>
#include <GL/gl.h>
#include <GL/glu.h>

using namespace std;
using namespace pcrepp;

struct Triangle
{
	Vec3f* verts[3];
	Vec3f* norms[3];
	Material *mtl;
	
	Triangle(Vec3f* a, Vec3f* na, Vec3f* b, Vec3f* nb, Vec3f* c, Vec3f* nc, Material *m)
	{
		verts[0] = a; norms[0] = na;
		verts[1] = b; norms[1] = nb;
		verts[2] = c; norms[2] = nc;
		mtl = m;
	}
};


class Mesh
{
public:
	Mesh();
	Mesh(char *file);
	~Mesh();
	
	void parse(char* file);
	void parseMaterials(string f);
	
	void render();
	
	const BoundingBox *getBoundingBox() const {return bbox;}
	const vector<Vec3f*>& getVertices() const {return verts;}
	const vector<Triangle>& getTriangles() const {return triangles;}

private:
	
	//regexs for face declarations
	static Pcre regex_face_simple;
	static Pcre regex_face_full;
	
	BoundingBox *bbox;
	vector<Vec3f*> verts;
	vector<Vec3f*> normals;
	vector<Vec3f*> c_normals; //computed normals
	vector<Triangle> triangles;
	map<string, Material *> materials;
};

#endif
