// specular.fs
//
// per-pixel specular lighting

varying vec3 N, L;
varying float specularExp;

uniform sampler2D sampler0;

uniform bool color_toggle;
uniform bool toon_toggle;
uniform bool lighting_toggle;

void main(void)
{
	//get SSAO value
	float ao = texture2D(sampler0, gl_FragCoord.xy/512.0).r;
	
    vec3 NN = normalize(N);
    vec3 NL = normalize(L);
    vec3 NH = normalize(NL + vec3(0.0, 0.0, 1.0));

    // calculate diffuse lighting
    float intensity = (max(0.0, dot(NN, NL)) + 0.1) * ao + 0.2;
    
    vec3 diffuse = gl_Color.rgb * intensity;

    // calculate specular lighting
    vec3 specular = vec3(0.0);
    if (intensity >= 0.0 && specularExp > 1.0)
    {
        intensity += pow(max(0.0, dot(NN, NH)), specularExp);
    }
    
    if (!lighting_toggle)
    {
    	intensity = ao;
    }

	if (toon_toggle)
	{
		intensity = floor(intensity * 4.0) / 4.0;
	}

    // sum the diffuse and specular components
    if (color_toggle)
    {
    	gl_FragColor.rgb = gl_Color.rgb * intensity;
    }
    else
    {
    	gl_FragColor.rgb = vec3(intensity);
    }
    
    gl_FragColor.a = gl_Color.a;
}
