// ssao.fs
//
// Screen Space Ambient Occlusion

uniform vec3 sample_points[8];

uniform sampler2D s_depth;
uniform sampler2D s_norm;

#define dp 1.0/512

const float radius = 16 * dp;

void main(void)
{
	//get the depth
    vec4 f_depth = texture2D(s_depth, gl_TexCoord[0].st);
    vec3 f_norm = normalize(texture2D(s_norm, gl_TexCoord[1].st).xyz);
        
    float occlusion = 0.0;
   
    float depth_sample;
    vec2 sp;
    for(int i=0; i < 8; i++)
    {
    	sp = radius * (reflect(sample_points[i], f_norm).xy) + gl_TexCoord[0].st;
    	depth_sample = f_depth.r - texture2D(s_depth, sp).r;
		
		//if the sample is closer than the current pixel -> occlude
		if(depth_sample > 0)// && depth_sample < radius * 0.05)
		{
			occlusion += 1.0;
		}
    }
	
	if(occlusion < 5.0){
		occlusion = 0;
	}else{
		//quadratic mapping
		occlusion = pow(occlusion, 2.0) / 64.0;
	}
	
	//cubic mapping
	//occlusion = pow(occlusion,3.0) / 512.0;

	//linear mapping
	//occlusion /= 8.0;
	
	//output shade
	gl_FragColor.rgb = vec3(1.0-occlusion);
}
