// toon_outlining.fs
//
// Depth discontinuities for toon outlining

vec2 offsets [8];

uniform sampler2D s_depth;
uniform sampler2D s_norm;

void main(void)
{
	vec4 depth = texture2D(s_depth, gl_TexCoord[0].st);
	vec4 norm = texture2D(s_norm, gl_TexCoord[0].st);
	float dp = 1.0 / 512.0;

	offsets[0] = vec2(-dp,-dp);
	offsets[1] = vec2(-dp,0);
	offsets[2] = vec2(-dp,dp);
	
	offsets[3] = vec2(0,-dp);
	offsets[4] = vec2(0,dp);
	
	offsets[5] = vec2(dp,-dp);
	offsets[6] = vec2(dp,0);
	offsets[7] = vec2(dp,dp);
	
	float darkness_depth = 0.0;
	float darkness_norm = 0.0;
	
    float base_depth = depth.r;
    vec3 base_norm = normalize(norm.xyz);
    
    float threshhold_depth = 0.05 * (1.0 - depth.r);
    float threshhold_norm = 0.95;
    
    for(int i = 0; i < 8; i++)
    {
    	depth = texture2D(s_depth, gl_TexCoord[0].st + offsets[i]);
    	norm.xyz = normalize(texture2D(s_norm, gl_TexCoord[0].st + offsets[i]).xyz);
    	
    	if (abs(depth.r - base_depth) > threshhold_depth) darkness_depth += 1.0;
    	if (dot(norm.xyz, base_norm) < threshhold_norm) darkness_norm += 1.0;
    }
	
	float illum = 1.0 - (darkness_depth + darkness_norm);

    gl_FragColor.rgb = vec3(illum, illum, illum);
    gl_FragColor.a = 1.0;
}
